"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
import numpy as np, os
import time

from numba import cuda

from inspectnn.BaseLayer import BaseLayer
from inspectnn.Conv.Conv_kernel_tflite import convolve_tflite_occorenza,convolve_tflite, convolve_tflite_padding,convolve_deepwise_padding,convolve_deepwise,convolve_tflite_mf, convolve_tflite_padding_mf,convolve_tflite_fake_mf, convolve_tflite_mfiber, convolve_tflite_fake_mfiber, convolve_tflite_padding_mfiber,convolve_tflite_fake #, convolve_tflite_mfiber_cpu
from inspectnn.Conv.ConvLayer import ConvLayer
from inspectnn.Conv.Conv_cuda import Conv_cuda


class ConvLayer_TFLITE(ConvLayer):
    total_type_layer_time = 0
    def __init__(self, stride = (1, 1), padding = (0,0),padding_type='valid', activation = "relu", quant_nbits = 8, multiplier = BaseLayer.default_multiplier, name = "Conv_TFlite",offset=[128,128],print_output=None,DEPTHWISE_CONV_2D=False):
        super().__init__(stride=stride,activation=activation, quant_nbits=quant_nbits, multiplier=multiplier, name=name,print_output=print_output,padding=padding,offset=offset)
        self.padding_type = padding_type
        self.idx_multiply = None
        self.mul_per_fibers = None
        self.preA = None
        self.DEPTHWISE_CONV_2D = DEPTHWISE_CONV_2D
       
        

    def __deepcopy__(self, memo = None):
        return ConvLayer_TFLITE(stride = self.stride, padding = self.padding, activation = self.activation, quant_nbits = self.quant_nbits, multiplier = self.multiplier, name = self.name)

    #@jit
    def forward_pass_input(self, inputv):
        #(np.sum(A_global_mem.copy_to_host()[7:14,7:14]*self.kernel_global_mem.copy_to_host()[k])*self.k_mul.copy_to_host()[k] + self.k_bias.copy_to_host()[k] - self.weights.copy_to_host()[k])*self.output_mul+self.output_offset
        A_global_mem = inputv
        #print(self.input_shape)
        self.test_cuda.setPointerInput(A_global_mem.device_ctypes_pointer.value,
                                      self.activation)
       

        #TODO: rinominare M
        #print(self.name,'dati input',A_global_mem.copy_to_host()[0])
        

        if(self.idx_multiply is None and self.mul_per_fibers is None):

            if (self.normal_kernel):
                
                #?direct CUDA conv
                self.test_cuda.convolutionRowsGPU()
                #cuda.synchronize()
                #convolve_tflite[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.M,
                #                                             self.activation,self.output_mul,self.output_offset,self.stride,self.update_bias,self.kp)
            elif (self.fake_kernel):
                
                #?direct CUDA conv
                self.test_cuda.convolutionRowsGPU_fake()
                #cuda.synchronize()
                #convolve_tflite_fake[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.M,
                #                                             self.activation,self.output_mul,self.output_offset,self.stride,self.update_bias,self.kp)
            else:
                
                #value_null = self.pre_layer.output_offset
                #self.test_cuda.setStride(self.stride[0],self.stride[1],value_null)
                
                #?direct CUDA conv
                self.test_cuda.convolutionRowsGPU_padding()
                
                #?Non direct CUDA conv
                #cuda.synchronize()
                #convolve_tflite_padding[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.M,
                #                                                                self.activation,self.output_mul,self.output_offset,value_null,self.stride,self.update_bias,self.kp)
        #! Fiber-based approximation 
        elif(self.mul_per_fibers is not None):
            
            # print(cuda.to_device(self.all_M))
            # exit()
            
            idxm = cuda.to_device(np.ascontiguousarray(self.mul_per_fibers))
            if (self.normal_kernel):
                cuda.synchronize()

                
                #convolve_tflite_mfiber_cpu(self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
                #                                            self.activation,self.output_mul,self.output_offset,self.stride,idxm,self.kp)
                convolve_tflite_mfiber[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
                                                            self.activation,self.output_mul,self.output_offset,self.stride,idxm,self.kp)
            elif (self.fake_kernel):
                
                cuda.synchronize()
                convolve_tflite_fake_mfiber[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
                                                            self.activation,self.output_mul,self.output_offset,self.stride,idxm,self.kp)

            else:
                value_null = self.pre_layer.output_offset
                
                cuda.synchronize()
                convolve_tflite_padding_mfiber[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
                                                                                self.activation,self.output_mul,self.output_offset,value_null,self.stride,idxm,self.kp)   
        #! Filter-based approximation    
        else:
            
            idxm = cuda.to_device(np.ascontiguousarray(self.idx_multiply))
            if (self.normal_kernel):
                cuda.synchronize()
                convolve_tflite_mf[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
                                                            self.activation,self.output_mul,self.output_offset,self.stride,idxm,self.kp)
            elif (self.fake_kernel):
                cuda.synchronize()
                convolve_tflite_fake_mf[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
                                                            self.activation,self.output_mul,self.output_offset,self.stride,idxm,self.kp)
            else:
                value_null = self.pre_layer.output_offset
                cuda.synchronize()
                convolve_tflite_padding_mf[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
                                                                                self.activation,self.output_mul,self.output_offset,value_null,self.stride,idxm,self.kp)
        #else:
            # idxm = cuda.to_device(np.ascontiguousarray(self.idx_multiply))
            # if (self.normal_kernel):
            #     cuda.synchronize()
            #     convolve_tflite_fibre[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
            #                                                 self.activation,self.output_mul,self.output_offset,self.stride,idxm,self.kp)
            # elif (self.fake_kernel):
            #     cuda.synchronize()
            #     convolve_tflite_fake_mf[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
            #                                                 self.activation,self.output_mul,self.output_offset,self.stride,idxm,self.kp)
            # else:
            #     value_null = self.pre_layer.output_offset
            #     cuda.synchronize()
            #     convolve_tflite_padding_mf[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
            #                                                                     self.activation,self.output_mul,self.output_offset,value_null,self.stride,idxm,self.kp)
        
        
        #print(self.name,"out:\n",self.results.copy_to_host()[0][0])  
        
        #self.test_cuda.convolutionRowsGPU(self.results.device_ctypes_pointer.value)
        #TODO: calcolare grid_dim per quant 3d

        #print(self.name,"out",self.results.copy_to_host()[0][0])
        

        if self.gpu_output_memory == False:
            self.outputv[:,:,:] = self.results.copy_to_host()            
    #@jit
    def forward_pass(self):
        
        #TODO: rinominare variabili

        if self.preA != self.pre_layer.results:
            A_global_mem = self.pre_layer.results
            self.preA  = A_global_mem
            #self.test_cuda.setConvolutiontexSrc(A_global_mem.device_ctypes_pointer())
            #print(self.name,"Ok ####################################################")
            
            if(self.idx_multiply is None and self.mul_per_fibers is None):
                self.test_cuda.cuda_set(self, A_global_mem )
            
            if not self.fake_kernel and not self.normal_kernel:
                self.value_null2 = self.pre_layer.output_offset
                self.test_cuda.setStride(self.stride[0],self.stride[1], self.value_null2)
            

        else:
            A_global_mem = self.preA
        
        #print(self.name)#,"in",A_global_mem.copy_to_host()[0])      
           
        #TODO: rinominare M
        #! Layer-based approximation  
        if(self.idx_multiply is None and self.mul_per_fibers is None):
            #self.cuda_set( A_global_mem )
            
            
            if (self.DEPTHWISE_CONV_2D):
                if( self.padding_type == "same"):
                    value_null = self.pre_layer.output_offset
                    
                    #self.test_cuda.convolutionRowsGPU_deepwise()
                    cuda.synchronize()
                    #print("Valid padding")
                    
                    convolve_deepwise_padding[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.M,
                                                                                    self.activation,self.output_mul,self.output_offset,value_null,self.stride,self.update_bias)
                    #print(self.name,"out:\n",self.results.copy_to_host()[0][0])  
                else:
                    #print("No padding: ",self.padding_type)
                    #self.test_cuda.print_parameters()
                    #self.test_cuda.convolutionRowsGPU_deepwise()
                    #print(self.name,"out:\n",self.results.copy_to_host()[0][0])  
                    cuda.synchronize()
                    convolve_deepwise[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.M,
                                                                                    self.activation,self.output_mul,self.output_offset,self.stride,self.update_bias)
                    
                    #print(self.name,"out:\n",self.results.copy_to_host()[0][0])  
            elif(self.normal_kernel):
                
                #?direct CUDA conv
                #cuda.synchronize()
                self.test_cuda.convolutionRowsGPU()

                #?Non direct CUDA conv
                #cuda.synchronize()
                #convolve_tflite[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.M,
                #                                            self.activation,self.output_mul,self.output_offset,self.stride,self.update_bias,self.kp)
            elif (self.fake_kernel):
                #?direct CUDA conv
                self.test_cuda.convolutionRowsGPU_fake()

                #?Non direct CUDA conv
                #TODO:funzione in numba bugata
                #cuda.synchronize()
                #convolve_tflite_fake[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.M,
                #                                            self.activation,self.output_mul,self.output_offset,self.stride,self.update_bias,self.kp)
            
              
                
            else:
                value_null = self.pre_layer.output_offset

                #?direct CUDA conv
                self.test_cuda.convolutionRowsGPU_padding()
                
                #?Non direct CUDA conv
                cuda.synchronize()
                convolve_tflite_padding[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.M,
                                                                                self.activation,self.output_mul,self.output_offset,value_null,self.stride,self.update_bias,self.kp)

           
        #! Fiber-based approximation         
        elif(self.mul_per_fibers is not None):
            
        
            idxm = cuda.to_device(np.ascontiguousarray(self.mul_per_fibers))
            
            # print(cuda.to_device(self.all_M))
            # exit()
            if (self.normal_kernel):
                
                cuda.synchronize()
                
                
                #convolve_tflite_mfiber_cpu(self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
                #                                            self.activation,self.output_mul,self.output_offset,self.stride,idxm,self.kp)
                
                convolve_tflite_mfiber[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
                                                            self.activation,self.output_mul,self.output_offset,self.stride,idxm,self.kp)
     
            elif (self.fake_kernel):
                cuda.synchronize()
                convolve_tflite_fake_mfiber[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
                                                        self.activation,self.output_mul,self.output_offset,self.stride,idxm,self.kp)

            
            else:
                value_null = self.pre_layer.output_offset
                cuda.synchronize()
                convolve_tflite_padding_mfiber[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
                                                                                self.activation,self.output_mul,self.output_offset,value_null,self.stride,idxm,self.kp)
                    
        #! Filter-based approximation    
        else:
            
            idxm = cuda.to_device(np.ascontiguousarray(self.idx_multiply))
            if (self.normal_kernel):
                cuda.synchronize()
                convolve_tflite_mf[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
                                                            self.activation,self.output_mul,self.output_offset,self.stride,idxm,self.kp)
            elif (self.fake_kernel):
                cuda.synchronize()
               
                convolve_tflite_fake_mf[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
                                                            self.activation,self.output_mul,self.output_offset,self.stride,idxm,self.kp)
            else:
                value_null = self.pre_layer.output_offset
                cuda.synchronize()
                
                convolve_tflite_padding_mf[self.griddim, self.blockdim](self.results,A_global_mem,self.kernel_global_mem,self.params,self.k_mul,self.weights,self.k_bias,self.all_M,
                                                                                self.activation,self.output_mul,self.output_offset,value_null,self.stride,idxm,self.kp)
        
        if self.gpu_output_memory == False:
            self.outputv[:,:,:] = self.results.copy_to_host()          
    
        #print(self.name,"out:\n",self.results.copy_to_host()[0][0])      
        #exit()  
    
    def load_weights(self, **kwargs):
        #TODO: sfrutare anche super().load_weights(**kwargs)
        
        self.test_cuda = Conv_cuda()
        
        self.kp=np.int8(1)
        self.enable_gpu = kwargs["enable_gpu"] 
        self.input_shape= kwargs["input_shape"]
        self.weights_cpu=np.array(kwargs["weights"],dtype=np.int16)
        self.kernel_shape = np.shape(kwargs["weights"])        
        self.output_shape = kwargs["output_shape"]

        self.n_channels=self.kernel_shape[0]

        self.params = cuda.to_device([self.n_channels,128,128])
        self.n_filter = self.n_channels
        self.quanto_k = None
        self.quanto_b = None

        
        #TODO: considera che è sempre su gpu
        if self.enable_gpu:
            self.use_gpu=True
            self.M = self.multiplier
            #TODO: parametrizare il k
            self.blockdim = (2,2,4)
            self.griddim = (self.output_shape[0] // self.blockdim[0] + 1, self.output_shape[1] // self.blockdim[1] + 1,self.output_shape[2] // (self.blockdim[2]*self.kp) + 1)#,n_channels)
            self.occorenza = None#cuda.to_device(np.zeros((self.n_filter,256,256),dtype=np.int32))
            self.kernel_global_mem = cuda.to_device(np.array(kwargs["weights"],dtype=np.int8))
           
             
            
            self.results = cuda.device_array(self.output_shape, dtype=np.int8)
            self.results2 = cuda.device_array(self.output_shape, dtype=np.int32)

            self.gpu_input_memory = kwargs["gpu_input_memory"]

            if ( kwargs["quant_data"] is not None):
                self.quantization=True

                self.quanto_b = cuda.to_device(np.ascontiguousarray(kwargs["quant_bias"]))
                self.quanto_k = cuda.to_device(np.ascontiguousarray(kwargs["quant_data"]))

                self.update_bias = cuda.to_device(np.ascontiguousarray(np.zeros(self.n_filter)))
                
                self.k_mul = cuda.to_device(np.ascontiguousarray(np.double( kwargs["quant_data"])*kwargs["quant_input_k"],dtype=np.double))#TODO:parametrizare il 255
                if(self.DEPTHWISE_CONV_2D):
                    self.weights = cuda.to_device(np.ascontiguousarray(np.sum(kwargs["weights"],(0,1,2),dtype=np.double)*kwargs["quant_data"]*kwargs["quant_input_offset"]*kwargs["quant_input_k"],dtype=np.double))#TODO:parametrizare il diviso 2 con alfa input+offest input
                
                else:
                    self.weights = cuda.to_device(np.ascontiguousarray(np.sum(kwargs["weights"],(1,2,3),dtype=np.double)*kwargs["quant_data"]*kwargs["quant_input_offset"]*kwargs["quant_input_k"],dtype=np.double))#TODO:parametrizare il diviso 2 con alfa input+offest input
                
                self.k_bias = cuda.to_device(np.ascontiguousarray(kwargs["biases"]*kwargs["quant_bias"],dtype=np.double))

                self.output_mul= 1.0/kwargs["quant_output_k"]
                self.output_offset= kwargs["quant_output_offset"]

            if self.print_output:
                self.outputv = np.zeros(self.output_shape)              
        else:
            self.convkernel, self.biases =kwargs["weights"], kwargs["biases"]
            self.outputv = np.zeros(self.output_shape)
        
        
        self.test_cuda.setPointerData(None,
                                      self.results.device_ctypes_pointer.value,
                                      self.kernel_global_mem.device_ctypes_pointer.value,
                                      self.k_mul.device_ctypes_pointer.value,
                                      self.weights.device_ctypes_pointer.value,
                                      self.k_bias.device_ctypes_pointer.value,
                                      self.output_mul,self.output_offset)
                
        self.test_cuda.setSizeData(self.input_shape[0],self.input_shape[1],self.input_shape[2],self.output_shape[0],self.output_shape[1],self.output_shape[2],
                                   self.kernel_shape[0],self.kernel_shape[1],self.kernel_shape[2],self.kernel_shape[3],256)
           
        self.test_cuda.setStride(self.stride[0],self.stride[1],0)
        
        #TODO: valutare se spostare nella convLayerBase
        self.padding_kernel = (self.padding_type != 'valid')#mannagia
        self.fake_kernel = (kwargs["weights"].shape[1] == 1 and kwargs["weights"].shape[2] == 1)
        self.normal_kernel = not self.padding_kernel and not self.fake_kernel
        if(self.activation == "relu" or self.activation == "Relu"):
            
            self.activation = 1
        elif(self.activation == "relu6" or self.activation == "Relu6"):
            self.activation = 6
        else:
            self.activation = 0
            
        #if True:#None == self.gpu_input_memory:#TODO:parametrizare meglio
                #self.outputv = np.zeros(self.output_shape,dtype=np.int32)          
        return self.output_shape
  
    # def cuda_set(self,A_global_mem ):
       
    #     #self.test_cuda.setConvolutiontexSrc(A_global_mem.device_ctypes_pointer())
        
    #     self.test_cuda.setPointerData(A_global_mem.device_ctypes_pointer.value,
    #                                   self.results.device_ctypes_pointer.value,
    #                                   self.kernel_global_mem.device_ctypes_pointer.value,
    #                                   self.k_mul.device_ctypes_pointer.value,
    #                                   self.weights.device_ctypes_pointer.value,
    #                                   self.k_bias.device_ctypes_pointer.value,
    #                                   self.output_mul,self.output_offset)
                
    #     self.test_cuda.setSizeData(self.input_shape[0],self.input_shape[1],self.input_shape[2],self.output_shape[0],self.output_shape[1],self.output_shape[2],
    #                                self.kernel_shape[0],self.kernel_shape[1],self.kernel_shape[2],self.kernel_shape[3],256)
                                   
    #     self.test_cuda.setPointerInput(A_global_mem.device_ctypes_pointer.value,
    #                                   self.activation)

    #     self.test_cuda.setPointerMul(self.M.device_ctypes_pointer.value)
        
    #     self.test_cuda.setStride(self.stride[0],self.stride[1],self.pre_layer.output_offset)
    def load_input_layer(self,layers,tensors_out):
        
        super().load_input_layer(layers,tensors_out)
        self.test_cuda.cuda_set( self,None )
        if not self.fake_kernel and not self.normal_kernel:
                self.value_null2 = self.pre_layer.output_offset
                self.test_cuda.setStride(self.stride[0],self.stride[1], self.value_null2)
        
    def research_hyperparameters(self):
        
        best_blockdim = self.blockdim
        best_griddim = self.griddim 
        best_kp = self.kp
        best_time = 100000000
        for i in range(0,16):
            for j in range(0,10):
                for k in range(0,5):
                    for kp in range(1,5):
                        if k*kp <=self.output_shape[2]:
                            self.kp = kp
                            self.blockdim = (i+1,j+1,k+1)
                            self.griddim = (self.output_shape[0] // self.blockdim[0] + 1, self.output_shape[1] // self.blockdim[1] + 1,self.output_shape[2] // (self.blockdim[2]*kp) + 1)
                            init_time = time.time()
                            
                            for asd in range(2):
                                self.forward_pass()
                                
                            tot_time = time.time() - init_time
                            #print("blockdim:",self.blockdim,"griddim:",self.griddim,"time:",tot_time)
                            if tot_time < best_time:
                                best_blockdim = self.blockdim
                                best_griddim = self.griddim
                                best_time = tot_time
                                best_kp = self.kp
                
        print("best_blockdim:",best_blockdim,"best_griddim:",best_griddim,"kp:",best_kp,"time:",best_time)
        self.kp = best_kp
        self.blockdim = best_blockdim
        self.griddim = best_griddim


"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
from numba import cuda, int8, int32
import time
TPB=16 #TODO:calcolare il tpb dinamicamente   

@cuda.jit(cache=True)
def mean(R, inputv, stride, pool_size,k_p):#, pooling):
    x, y,k = cuda.grid(3)
    if x < R.shape[0] and y < R.shape[1] and k*k_p < R.shape[2]:
        x_start = x * stride[0]
        x_end = min(x_start + pool_size[0],inputv.shape[0])
        y_start = y * stride[1]
        y_end = min(y_start + pool_size[1],inputv.shape[1])
        
        temp_max = cuda.local.array(shape=(16), dtype=int32)
        
        k = k * k_p
        k_max = min(R.shape[2],k+k_p)
        for p in range (k,k_max):
            temp_max[p-k]=-90000
            
        for i in range(x_start,x_end):
            for j in range(y_start,y_end):
                #for p in range (k,k_max):
                idp= 0
                for p in range (k,k_max):
                    
                    temp_max[idp]=max(inputv[i,j,p],temp_max[idp])
                    idp+=1
                        
        for p in range (k,k_max):                
            R[x, y, p] = temp_max[p-k]
    


#TODO: gestire il padding del pool avg
@cuda.jit(cache=True)
def mean_avg(R, inputv, stride, pool_size,k_p):#, pooling):
    x, y,k = cuda.grid(3)
    if x < R.shape[0] and y < R.shape[1] and k*k_p < R.shape[2]:
        x_start = x * stride[0]
        x_end = x_start + pool_size[0]
        y_start = y * stride[1]
        y_end = y_start + pool_size[1]
        temp_sum = cuda.local.array(shape=(16), dtype=int32)
        #temp_sum = 0
        k = k * k_p
        k_max = min(R.shape[2],k+k_p)
        for p in range (k,k_max):
            temp_sum[p-k]=0
            for i in range(x_start,x_end):
                for j in range(y_start,y_end):
                    
                    temp_sum[p-k]+=inputv[i,j,p]
        
            R[x, y, p] = round(temp_sum[p-k]/(pool_size[0]*pool_size[1]))
        
        
        
 


"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
from numba import cuda, int8

@cuda.jit(cache=True)
def quant_3d_FakeMinMax(results, values,out_mul,out_offest,in_mul,in_offset):
    """
    Find the maximum value in values and store in result[0].
    Both result and values are 3d arrays.
    """
   
    #TODO: valutare l' else 8risultato con matrice compresa con solo o 0 o val max
    i, j,k = cuda.grid(3)
    
    max_i=results.shape[0]
    max_j=results.shape[1]
    max_k=results.shape[2]
    
    
    d_i = (results.shape[0]-values.shape[0])/2
    d_j = (results.shape[1]-values.shape[1])/2
    
    
    #if the thread coordinates are outside of the image, we ignore the thread:
    if ((i >= max_i)  or (j >= max_j) or (k >= max_k)): 
        return

    #results[i,j,k] = round(values[i,j,k]*out_mul + out_offest)
    if i-d_i<0 or j-d_j<0 or i-d_i>=values.shape[0] or j-d_j>=values.shape[1] :
        results[i,j,k] = out_offest
    else:
        results[i,j,k] = values[i-d_i,j-d_j,k] #-in_offset)//in_mul*out_mul + out_offest)
    #results[i,j,k] = round(2**(quant_nbits+1)*(values[i,j,k]- 2**(quant_nbits)-t_min)/(t_max-2*t_min))#TODO:meno -1
    # CUDA conv kernel


@cuda.jit(cache=True)
def pad_3d(results, values,out_offest):
    i, j,k = cuda.grid(3)
    
    max_i=results.shape[0]
    max_j=results.shape[1]
    max_k=results.shape[2]
    
    #if the thread coordinates are outside of the image, we ignore the thread:
    if ((i >= max_i)  or (j >= max_j) or (k >= max_k)): 
        return
    
    d_i = (results.shape[0]-values.shape[0])//2
    d_j = (results.shape[1]-values.shape[1])//2
    
    #results[i,j,k] = round(values[i,j,k]*out_mul + out_offest)
    if i-d_i<0 or j-d_j<0 or i-d_i>=values.shape[0] or j-d_j>=values.shape[1] :
        results[i,j,k] = out_offest
    else:
        results[i,j,k] = values[i-d_i,j-d_j,k] #-in_offset)//in_mul*out_mul + out_offest)


@cuda.jit(cache=True)
def quant_2d(results, values,out_mul,out_offest,in_mul,in_offset):
    """
    Find the maximum value in values and store in result[0].
    Both result and values are 3d arrays.
    """

    i, j = cuda.grid(2)
    
    max_i=results.shape[0]
    max_j=results.shape[1]

    
    #if the thread coordinates are outside of the image, we ignore the thread:
    if ((i >= max_i)  or (j >= max_j) ): 
        return

    #TODO:fare un unico coeficente out_mul/in_mul
    results[i,j] = round((values[i,j]-in_offset)*out_mul/in_mul + out_offest)
    


try:
    import setuptools
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from codecs import open
from os import path

setup(
    name="inspectnn",
    version="0.6.1",
    description="Inference eNgine uSing aPproximate arithmEtic ComponenTs for Neural Networks",
    long_description="Inference eNgine uSing aPproximate arithmEtic ComponenTs for Neural Networks",
    url="https://github.com/SalvatoreBarone/inspect-nn",
    author="Salvatore Barone, Filippo Ferrandino",
    author_email="salvatore.barone@unina.it, filippo.ferrandino@unina.it",
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3.11"
    ],
    keywords="Neural Networks Approximate-Computing",
    packages=setuptools.find_packages(),#["inspectnn"],
    package_data={
        "": ["**.so"],  
    },
    include_package_data=True,

    python_requires='>=3.11',#3.11 o 3.12
    #numba 0.57.0rc1 va con python 3.11 e numpy 1.24                   numba==0.59.0rc1
    install_requires=["tensorflow>=2.13.0", "protobuf","numpy>=1.23","numba>=0.58.0","scikit-learn"],#fare -U pymoo #"scipy==1.10.1"
    project_urls={
        "Bug Reports": "https://github.com/SalvatoreBarone/inspect-nn/issues",
        "Source": "https://github.com/SalvatoreBarone/inspect-nn",
    },
    

)




from sklearn.model_selection import train_test_split
import numpy as np, tensorflow as tf, random
from sklearn.model_selection import train_test_split

import pandas as pd
import os,pandas

from tensorflow.keras.preprocessing.image import ImageDataGenerator
from sklearn.utils import shuffle



IMAGE_HEIGHT = 224
IMAGE_WIDTH = 224
IMAGE_CHANNELS = 3

my_PATH = os.path.dirname(__file__)+ '/FGVC7/train.csv'



def get_class(row):
    
    if row['multiple_diseases'] == 1:
        return 'multiple_diseases'
    
    elif row['rust'] == 1:
        return 'rust'
    
    elif row['scab'] == 1:
        return 'scab'
    
    else:
        return 'healthy'

     
def Dataset_FGVC7(num_of_samples,sequenzial=False,PATH=None):
    
        
    
    df_train_all = pd.read_csv(PATH+"train.csv")
    
    df_train_all['target'] = df_train_all.apply(get_class, axis=1)

    df_train_all.head()
    df_train_all['target'].value_counts()
    
    # shuffle
    df_train_all_shuffle = shuffle(df_train_all, random_state=101)
    # select the column that we will use for stratification
    y = df_train_all_shuffle['target']

    df_train, df_val = train_test_split(df_train_all_shuffle, test_size=0.2, random_state=101, stratify=y)
    

    
    animals = os.listdir(PATH)
    
    images = [os.path.join(PATH, "images/"+images_id +".jpg") for images_id in df_train_all['image_id']]
    
    label = []
    categories = []
    for animal in df_train_all['target']:
        if "rust" in animal:
            categories.append('2s')
        elif "scab" in animal:
            categories.append('3')

        elif "healthy" in animal:
            categories.append('0')
        elif "multiple_diseases" in animal:
            categories.append('1')
        else:
            print(animal + " non identificato!")
            
    animals = pandas.DataFrame({'Image_name': np.array([f'{f}' for f in images]), 'Category': categories})
    
    if num_of_samples > 0:
        batch_size = num_of_samples
    else:
        batch_size = 1-num_of_samples
        
        
    test_datagen = ImageDataGenerator(rescale=1./255)
    test_generator = test_datagen.flow_from_dataframe(animals, x_col='Image_name', y_col='Category', class_mode = 'sparse', target_size = (224, 224), batch_size = batch_size, validate_filenames = False,shuffle=not sequenzial)
    y_test = animals['Category']
    len_test_dataset = animals.shape[0]
    
  
    # for i in range(len(test_generator)):
    #    batch_images, batch_labels = test_generator[i]
    #    for j in range(batch_images.shape[0]):
    #        image = batch_images[j]
    #        label = batch_labels[j]

    if(num_of_samples > 0):

        img, lab = test_generator[0]
        #print(lab)
        #for i in range(1,num_of_samples):
        #    #trasforma img2 da bgr a rgb senza utilizare cv
        #    img = img2[i][:,:,::-1]
        
        label = [int(j) for j in lab]
        return  None,label ,img, lab
    
    
    else:
        batch_images, batch_labels = test_generator[0]
        
        label = [int(batch_labels[-num_of_samples])]
        return  None,label,[batch_images[-num_of_samples]], [batch_labels[-num_of_samples]]


"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
import numpy as np, os
import time

from numba import cuda

from inspectnn.BaseLayer import BaseLayer
from inspectnn.Conv.Conv_kernel_tflite import convolve_tflite_occorenza,convolve_tflite, convolve_tflite_padding,convolve_tflite_fake,convolve_tflite_mf, convolve_tflite_padding_mf,convolve_tflite_fake_mf
from inspectnn.Conv.ConvLayer import ConvLayer

import ctypes

class Conv_cuda(object):
    def __init__(self, size_H=3, SIZE_W=3, size_filter=3):
        dir_path = os.path.dirname(os.path.abspath(__file__))
        
        self.lib = ctypes.cdll.LoadLibrary(f'{dir_path}/libconvolutionTexture.so')
        #self.lib = ctypes.cdll.LoadLibrary('/home/ssaa/Git/inspect-nn/inspectnn/Conv/conv_cuda/convolutionTexture/libconvolutionTexture.so')
        self.size_H = size_H
        self.size_W = SIZE_W
        self.size_filter = size_filter

        # Declare input and output types for each method you intend to use
        #float *d_Dst, int imageW,int imageH, cudaTextureObject_t texSrc
        self.lib.convolutionRowsGPU.argtypes = [ctypes.c_void_p]
        self.lib.convolutionRowsGPU.restype = ctypes.c_void_p
        
        self.lib.convolutionRowsGPU_fake.argtypes = [ctypes.c_void_p]
        self.lib.convolutionRowsGPU_fake.restype =  ctypes.c_void_p
        
        self.lib.convolutionRowsGPU_padding.argtypes = [ctypes.c_void_p]
        self.lib.convolutionRowsGPU_padding.restype =  ctypes.c_void_p
        
        self.lib.convolutionRowsGPU_deepwise.argtypes = [ctypes.c_void_p]
        self.lib.convolutionRowsGPU_deepwise.restype = ctypes.c_void_p
        
        self.lib.convolutionRowsGPU_padding_deepwise.argtypes = [ctypes.c_void_p]
        self.lib.convolutionRowsGPU_padding_deepwise.restype =  ctypes.c_void_p
        
        #lib.DataConv
        self.lib.getStruct.argtypes = []
        self.lib.getStruct.restype =  ctypes.c_void_p
        self.obj = self.lib.getStruct()
        #self.obj = lib.convolutionRowsGPU(val)
        
        self.lib.setSizeData.argtypes = [ctypes.c_void_p,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int]
        self.lib.setSizeData.restype =  ctypes.c_void_p
        
        self.lib.setPointerMul.argtypes = [ctypes.c_void_p,ctypes.c_void_p]
        self.lib.setPointerMul.restype = ctypes.c_void_p
        
        
        self.lib.setPointerData.argtypes = [ctypes.c_void_p,ctypes.c_void_p,ctypes.c_void_p,ctypes.c_void_p,
                                                        ctypes.c_void_p,ctypes.c_void_p,ctypes.c_void_p,
                                                        ctypes.c_double,ctypes.c_int]
        self.lib.setPointerData.restype = ctypes.c_void_p
        
        self.lib.print_parameters.argtypes= [ctypes.c_void_p]
        
        self.lib.setStride.argtypes = [ctypes.c_void_p,ctypes.c_int,ctypes.c_int,ctypes.c_int]
        
        
        self.lib.setPointerInput.argtypes = [ctypes.c_void_p,ctypes.c_void_p,ctypes.c_char]

    def setPointerInput(self,Input,activation):
        #print("Activation: ",activation)
        self.lib.setPointerInput(self.obj,Input,activation)
        
        
    def setStride(self,x,y,val_null=0):
        self.lib.setStride(self.obj,x,y,val_null)
        
    def setSizeData(self,in_x,in_y,in_z,out_x, out_y, out_z, k_x, k_y, k_z, k_p, size_m):
        self.lib.setSizeData(self.obj,in_x,in_y,in_z,out_x, out_y, out_z, k_x, k_y, k_z, k_p, size_m)
        
        
    def setPointerData(self,Input,d_Dst, Kernel,k_mul,weights,k_bias,out_mul, out_offset):
        self.lib.setPointerData(self.obj,d_Dst, Input, Kernel,k_mul,weights,k_bias,out_mul, out_offset)
            
    def setPointerMul(self, M):
        self.lib.setPointerMul(self.obj,M)
        
        return
    
    def convolutionRowsGPU_fake(self):
        cuda.synchronize()
        self.lib.convolutionRowsGPU_fake(self.obj)
    
    def convolutionRowsGPU_padding(self):
        cuda.synchronize()
        self.lib.convolutionRowsGPU_padding(self.obj)
       
    def convolutionRowsGPU(self):
        cuda.synchronize()
        self.lib.convolutionRowsGPU(self.obj)
    
    def convolutionRowsGPU_deepwise(self):
        cuda.synchronize()
        self.lib.convolutionRowsGPU_deepwise(self.obj)
    def convolutionRowsGPU_padding_deepwise(self):
        cuda.synchronize()
        self.lib.convolutionRowsGPU_padding_deepwise(self.obj)
           
    def print_parameters(self):
        self.lib.print_parameters(self.obj)

    def cuda_set(self,classLayer,A_global_mem ):
       
        if A_global_mem is not None:
            self.setPointerData(A_global_mem.device_ctypes_pointer.value,
                                      classLayer.results.device_ctypes_pointer.value,
                                      classLayer.kernel_global_mem.device_ctypes_pointer.value,
                                      classLayer.k_mul.device_ctypes_pointer.value,
                                      classLayer.weights.device_ctypes_pointer.value,
                                      classLayer.k_bias.device_ctypes_pointer.value,
                                      classLayer.output_mul,classLayer.output_offset)
            
            self.setPointerInput(A_global_mem.device_ctypes_pointer.value,
                                      classLayer.activation)
            self.setPointerMul(classLayer.M.device_ctypes_pointer.value)
            
        self.setSizeData(classLayer.input_shape[0],classLayer.input_shape[1],classLayer.input_shape[2],classLayer.output_shape[0],classLayer.output_shape[1],classLayer.output_shape[2],
                                   classLayer.kernel_shape[0],classLayer.kernel_shape[1],classLayer.kernel_shape[2],classLayer.kernel_shape[3],256)

        self.setStride(classLayer.stride[0],classLayer.stride[1],classLayer.pre_layer.output_offset)
        
    


import  ctypes,os
from numba import cuda

class Pool_cuda(object):
    def __init__(self, size_H=3, SIZE_W=3, size_filter=3):
        dir_path = os.path.dirname(os.path.abspath(__file__))
        
        self.lib = ctypes.cdll.LoadLibrary(f'{dir_path}/poolMul.so')
        #self.lib = ctypes.cdll.LoadLibrary('/home/ssaa/Git/inspect-nn/inspectnn/Conv/conv_cuda/convolutionTexture/libconvolutionTexture.so')
        self.size_H = size_H
        self.size_W = SIZE_W

        # Declare input and output types for each method you intend to use
        #float *d_Dst, int imageW,int imageH, cudaTextureObject_t texSrc
        

        self.lib.PoolGPU.argtypes = [ctypes.c_void_p]
        self.lib.PoolGPU.restype =  ctypes.c_void_p
        

        #lib.DataConv
        self.lib.getStruct.argtypes = []
        self.lib.getStruct.restype =  ctypes.c_void_p
        self.obj = self.lib.getStruct()
        #self.obj = lib.convolutionRowsGPU(val)
        
        self.lib.setSizeData.argtypes = [ctypes.c_void_p,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int]
        self.lib.setSizeData.restype =  ctypes.c_void_p
        
        self.lib.setStride.argtypes = [ctypes.c_void_p,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int]
        self.lib.setStride.restype =  ctypes.c_void_p
        
        self.lib.setPointerData.argtypes = [ctypes.c_void_p,ctypes.c_void_p,ctypes.c_void_p, ctypes.c_float,ctypes.c_int]
        self.lib.setPointerData.restype = ctypes.c_void_p
        
        self.lib.print_parameters.argtypes= [ctypes.c_void_p]
        
        self.lib.setPointerInput.argtypes = [ctypes.c_void_p,ctypes.c_void_p,ctypes.c_bool]

    def setPointerInput(self,Input,activation):
        #print("Activation: ",activation)
        self.lib.setPointerInput(self.obj,Input,activation)
        
        
    def setStride(self,x,y,sizex,sizey):
        self.lib.setStride(self.obj,x,y,sizex,sizey)
        
    def setSizeData(self,in_x,in_y,in_z,out_x,out_y, out_z):
        self.lib.setSizeData(self.obj,in_x,in_y,in_z,out_x,out_y, out_z)
        
        
    def setPointerData(self,Input,d_Dst, out_mul, out_offset):
        self.lib.setPointerData(self.obj,d_Dst, Input,out_mul, out_offset)
            
    def PoolGPU(self):
        cuda.synchronize()
        self.lib.PoolGPU(self.obj)
        
         
    def setMatrixKernel(self,kernel):
        self.lib.setMatrixKernel(kernel)
    
    def setConvolutiontexSrc(self,ASrc):
        self.lib.setConvolutiontexSrc(ASrc)
        
    def setConvolutiontexSrcv2(self,ASrc):
        self.lib.setConvolutiontexSrcv2(ASrc)
        


    def print_parameters(self):
        self.lib.print_parameters(self.obj)
        
        
    def cuda_set(self,classLayer,A_global_mem ):
       
        self.setPointerData(A_global_mem.device_ctypes_pointer.value,
                                      classLayer.results.device_ctypes_pointer.value,
                                      classLayer.output_mul,classLayer.output_offset)
        
        if classLayer._2d:
            self.setSizeData(classLayer.input_shape[0],classLayer.input_shape[1],classLayer.input_shape[2],classLayer.output_shape[0],classLayer.output_shape[1],0)
        else:
            self.setSizeData(classLayer.input_shape[0],classLayer.input_shape[1],classLayer.input_shape[2],classLayer.output_shape[0],classLayer.output_shape[1],classLayer.output_shape[2])
                                   
        self.setPointerInput(A_global_mem.device_ctypes_pointer.value,
                                      classLayer.activation)

        self.setStride(classLayer.stride[0],classLayer.stride[1],classLayer.pool_size[0],classLayer.pool_size[1])


"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
import numpy as np
from numba import cuda,jit
from inspectnn.BaseLayer import BaseLayer
from inspectnn.Add.Add_kernel_tflite import add3d_tflite,add3dv2_tflite
import os

import ctypes

class Test_cuda(object):
    def __init__(self, size_H=3, SIZE_W=3, size_filter=3):
        dir_path = os.path.dirname(os.path.abspath(__file__))
        
        self.lib = ctypes.cdll.LoadLibrary(f'{dir_path}/libAddTexture.so')
        #self.lib = ctypes.cdll.LoadLibrary('/home/ssaa/Git/inspect-nn/inspectnn/Conv/conv_cuda/convolutionTexture/libconvolutionTexture.so')
        self.size_H = size_H
        self.size_W = SIZE_W
        self.size_filter = size_filter

        self.lib.cuda_add.argtypes = [ctypes.c_void_p]
        self.lib.cuda_add.restype = ctypes.c_void_p
        

        #lib.DataConv
        self.lib.getStruct.argtypes = []
        self.lib.getStruct.restype =  ctypes.c_void_p
        self.obj = self.lib.getStruct()
        #self.obj = lib.convolutionRowsGPU(val)
        
        self.lib.setSizeData.argtypes = [ctypes.c_void_p,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int]
        self.lib.setSizeData.restype =  ctypes.c_void_p
        
        self.lib.setPointerMul.argtypes = [ctypes.c_void_p,ctypes.c_void_p]
        self.lib.setPointerMul.restype = ctypes.c_void_p
        
        
        self.lib.setPointerData.argtypes = [ctypes.c_void_p,ctypes.c_void_p,ctypes.c_void_p,ctypes.c_void_p,
                                                        ctypes.c_void_p,ctypes.c_void_p,ctypes.c_void_p,
                                                        ctypes.c_float,ctypes.c_int]
        self.lib.setPointerData.restype = ctypes.c_void_p
        
        self.lib.print_parameters.argtypes= [ctypes.c_void_p]
        
        self.lib.setData.argtypes = [ctypes.c_void_p,ctypes.c_float,ctypes.c_float,ctypes.c_float,ctypes.c_float]
        
        
        self.lib.setPointerInput.argtypes = [ctypes.c_void_p,ctypes.c_void_p,ctypes.c_bool]

    def setPointerInput(self,Input,activation):
        self.lib.setPointerInput(self.obj,Input,activation)
        
        
    def setData(self,of_A,k_A,of_B,k_B):
        self.lib.setData(self.obj,of_A,k_A,of_B,k_B)
        
    def setSizeData(self,in_x,in_y,in_z,out_x, out_y, out_z, k_x, k_y, k_z, k_p, size_m):
        self.lib.setSizeData(self.obj,in_x,in_y,in_z,out_x, out_y, out_z, k_x, k_y, k_z, k_p, size_m)
        
        
    def setPointerData(self,Input,d_Dst, Kernel,k_mul,k_filtri,k_bias,out_mul, out_offset):
        self.lib.setPointerData(self.obj,d_Dst, Input, Kernel,k_mul,k_filtri,k_bias,out_mul, out_offset)
            
    def setPointerMul(self, M):
        self.lib.setPointerMul(self.obj,M)
        
        return
    
    def cuda_add(self):
        cuda.synchronize()
        self.lib.cuda_add(self.obj)
         
    def setConvolutionKernel(self,kernel):
        self.lib.setConvolutionKernel(kernel)
    
    def print_parameters(self):
        self.lib.print_parameters(self.obj)


class AddLayer(BaseLayer):
    layer_name = "Add"
    def __init__(self, name = "cocatenate",activation = ""):
        super().__init__(self, name = name)
        self.activation = (activation == "relu" or activation == "Relu")

    def __deepcopy__(self, memo = None):
        return AddLayer(name = self.name)

    def load_weights(self, **kwargs):
        self.input_shape_A,self.input_shape_B, self.enable_gpu = kwargs["input_shape_A"],kwargs["input_shape_B"], kwargs["enable_gpu"] 
        self.output_shape = kwargs["output_shape"]
        self.outputv = np.zeros(self.output_shape)

        self.blockdim = (2,2,8)
        self.stride = 1
        self.griddim = (self.output_shape[0] // self.blockdim[0] + 1, self.output_shape[1] // self.blockdim[1] + 1,self.output_shape[2] // (self.stride*self.blockdim[2]) + 1)#,n_channels)

        self.test_cuda = Test_cuda()
        
        
        
        if self.enable_gpu:
            self.use_gpu = True
           
            self.results = cuda.device_array(self.output_shape, dtype=np.int8)
            self.gpu_layer_A = None
            self.gpu_layer_B = None

            self.output_K = kwargs["quant_output_k"]
            self.output_mul = 1/kwargs["quant_output_k"]
            self.output_offset = kwargs["quant_output_offset"]
        return self.output_shape
    
    #@jit
    def forward_pass(self):
        if self.gpu_layer_A is None:
            print("add no buoeno")

        else:
            
            
            
            self.test_cuda.cuda_add()
            #cuda.synchronize()
            #add3d_tflite[self.griddim, self.blockdim](self.results,self.gpu_layer_A.results,self.gpu_layer_B.results,
            #                                         self.A_mul,self.B_mul, self.A_off,self.B_off,
            #                                         self.activation,self.output_mul,self.output_offset)


            

    def load_input_layer(self,layers,tensors_out):
        op_in1=self.code_tensor_inputs[0]
        op_in2=self.code_tensor_inputs[1]

        self.gpu_layer_A = layers[tensors_out.index(op_in1)]
        self.gpu_layer_B = layers[tensors_out.index(op_in2)]

        self.gpu_output_memory = True

        self.A_mul=1/self.gpu_layer_A.output_mul
        self.B_mul=1/self.gpu_layer_B.output_mul

        self.A_off=self.gpu_layer_A.output_offset*1.0
        self.B_off=self.gpu_layer_B.output_offset*1.0
        
        self.cuda_set()
        
        
        
      
    def cuda_set(self ):
       
        #self.test_cuda.setConvolutiontexSrc(A_global_mem.device_ctypes_pointer())
        
        self.test_cuda.setPointerData(self.gpu_layer_A.results.device_ctypes_pointer.value,
                                      self.results.device_ctypes_pointer.value,
                                      self.gpu_layer_B.results.device_ctypes_pointer.value,
                                      None,None,None,
                                      self.output_mul,self.output_offset)
                
        self.test_cuda.setSizeData(self.input_shape_A[0],self.input_shape_A[1],self.input_shape_A[2],self.output_shape[0],self.output_shape[1],self.output_shape[2],
                                   self.input_shape_B[0],self.input_shape_B[1],self.input_shape_B[2],0,256)
                                   
        self.test_cuda.setPointerInput(self.gpu_layer_A.results.device_ctypes_pointer.value,
                                      self.activation)

        self.test_cuda.setData(self.A_off,self.A_mul,self.B_off,self.B_mul)

        


"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
from numba import cuda, jit
import numpy as np, json, copy, time
from inspectnn.Model.NetworkModel import NetworkModel 
from distutils.dir_util import mkpath
from inspectnn.Conv.ConvLayer import ConvLayer
from inspectnn.Dense.DenseLayer import DenseLayer
from inspectnn.Pooling.PoolingLayer_TFLITE import PoolingLayer_TFLITE
from inspectnn.Flatten.FlattenLayer import FlattenLayer
from inspectnn.Quantizate.QuantizateLayer import QuantizateLayer
from inspectnn.Add.AddLayer import AddLayer
from inspectnn.SoftmaxLayer import SoftmaxLayer
#from .OutputLayer import OutputLayer
from inspectnn.Utils.utils import grouped, flatten

class NetworkModelTflite(NetworkModel):
    #TODO:ereditare da NetworkModel
    def __init__(self, learned_parameters, layers,input_shape,print_input=False,quant_data=None,quant_data_scale=None):
        self.learned_parameters = learned_parameters
        self.layers = layers
        self.outputs = []
        self.input_shape=input_shape
        self.print_input = print_input
        self.total_time = 0
        self.n_frame_inference = 0
        self.elapsed_time = np.zeros(len(layers))
        self.quant_data=quant_data
        self.get_Numbers_layer_with_multiply()
        self.images_on_gpu=False
        self.init_quantizate = True
        self.n_filter = None

    def __deepcopy__(self, memo = None):
        return NetworkModelTflite(self.learned_parameters, copy.deepcopy(self.layers),self.input_shape)
   
    def predict(self, test_image):
        
        
        self.n_frame_inference+=1
       
        id_layer = 0
        id_start_layer = 1
        if self.init_quantizate:
            
            st = time.time()
            #print("Predict no pass:", test_image.copy_to_host()[0][0])
            self.layers[id_layer].forward_pass_input(inputv = test_image)
            
            et = time.time()
            self.elapsed_time[id_layer] += et - st
            type(self.layers[id_layer]).total_type_layer_time+= et - st
            self.total_time += et - st
        else:
            st = time.time()
            #print("Predict pass:", test_image.copy_to_host()[0][0])
            self.layers[id_start_layer].forward_pass_input(inputv = test_image)
            
            et = time.time()
            self.elapsed_time[id_start_layer] += et - st
            type(self.layers[id_layer]).total_type_layer_time+= et - st
            self.total_time += et - st
            id_start_layer = 2

       # idx = 0

        for idx, id_layer in enumerate(range(id_start_layer,len(self.layers)-1)):
            # if(idx == 2):
            #     exit()
            st = time.time()
            
            #print(id_layer,self.layers[id_layer].name)
            #if isinstance(self.layers[id_layer], (ConvLayer, DenseLayer,PoolingLayer_TFLITE, FlattenLayer,QuantizateLayer,AddLayer)):
            self.layers[id_layer].forward_pass()
            #else:
            #    print('Layer {:>15} not supported yet'.format(self.layers[id_layer].name))
            
            et = time.time()
            self.elapsed_time[id_layer] += et - st
            
            self.total_time += et - st
            type(self.layers[id_layer]).total_type_layer_time+= et - st
            #print(self.layers[id_layer].name,self.layers[id_layer].results.copy_to_host()[0][0])
            #idx+= 1
        return self.layers[id_layer].outputv
        
    def sfmpredict(self, test_image):
        outputv = self.predict(test_image)

        st = time.time()
        #print(self.name,self.blockdim,self.griddim)
        #print("layers: ",self.layers[id_layer].name,self.layers[id_layer].blockdim,self.layers[id_layer].griddim)
        if(self.print_input):
            self.layers[-1].update_save_path()
        if isinstance(self.layers[-1], (SoftmaxLayer)):
            self.layers[-1].forward_pass(outputv)
        else:
            print('Layer {:>15} not supported yet'.format(self.layers[-1].name))

        et = time.time()
        self.elapsed_time[-1] += et - st
        self.total_time += et - st

        return self.layers[-1].outputv
        
   

    


"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""

import numpy as np
from numba import cuda, jit
from inspectnn.BaseLayer import BaseLayer
from inspectnn.Mean.Mean_kernel import mean, mean_avg

class MeanLayer_TFLITE(BaseLayer):
    layer_name = "Mean"
    def __init__(self, stride = (1, 1), pool_size = (2, 2), pooling = "max", name = "layer", input_memory = None):
        super().__init__(self, name = name)
        self.stride = stride
        self.pool_size = pool_size
        self.pool = pooling

    def __deepcopy__(self, memo = None):
        return MeanLayer_TFLITE(stride = self.stride, pool_size = self.pool_size, pooling = self.pool, name = self.name)

    #@jit
    def forward_pass(self):

        A_global_mem=self.pre_layer.results
        avg = (self.pool == "avg" or self.pool == "Avg")

        if avg:
            cuda.synchronize()
            pool_avg[self.griddim, self.blockdim](self.results, A_global_mem,self.stride_global_mem, self.pool_size_global_mem,self.k_p)#, pooling)

        else:
            cuda.synchronize()
            pool[self.griddim, self.blockdim](self.results, A_global_mem,self.stride_global_mem, self.pool_size_global_mem,self.k_p)#, pooling)
        
        
        #print(self.name,"output:",self.results.copy_to_host()[0][3])
        #if self.gpu_output_memory == False:
            #self.outputv[:,:,:] =self.results.copy_to_host()


    def load_weights(self, **kwargs):
        self.input_shape, self.enable_gpu = kwargs["input_shape"],kwargs["enable_gpu"]
        #print(self.pool_size,self.stride)
        self.output_shape = [int(np.floor((self.input_shape[i] - self.pool_size[i]) / self.stride[i]) + 1) for i in range(2)] + [self.input_shape[2]]      
        self.outputv = np.zeros(self.output_shape)
        if self.enable_gpu:
            self.k_p=16
            self.blockdim = (4,4,8)
            self.griddim = (self.output_shape[0] // self.blockdim[0] + 1, self.output_shape[1] // self.blockdim[1] + 1,self.output_shape[2] // (self.blockdim[2]*self.k_p) + 1)#,n_channels)
            self.use_gpu=True

            self.results = cuda.device_array(self.output_shape, dtype=np.int8)
            self.stride_global_mem = cuda.to_device(self.stride)
            self.pool_size_global_mem = cuda.to_device(self.pool_size)
            self.gpu_input_memory = kwargs["gpu_input_memory"]

            self.output_mul= 1/kwargs["quant_output_k"]
            self.output_offset= kwargs["quant_output_offset"]
        return self.output_shape
    
    
    def research_hyperparameters(self):
        
        best_blockdim = self.blockdim
        best_griddim = self.griddim 
        best_kp = self.kp
        best_time = 100000000
        for i in range(0,4):
            for j in range(0,4):
                for k in range(1,8):
                    for kp in range(1,32):
                        if k*kp <= self.output_shape[2]:
                            self.k_p = kp
                            self.blockdim = (i+1,j+1,k+1)
                            self.griddim = (self.output_shape[0] // self.blockdim[0] + 1, self.output_shape[1] // self.blockdim[1] + 1,self.output_shape[2] // (self.blockdim[2]*kp) + 1)
                            init_time = time.time()
                            
                            for asd in range(2):
                                self.forward_pass()
                                
                            tot_time = time.time() - init_time
                            #print("blockdim:",self.blockdim,"griddim:",self.griddim,"time:",tot_time)
                            if tot_time < best_time:
                                best_blockdim = self.blockdim
                                best_griddim = self.griddim
                                best_time = tot_time
                                best_kp = self.k_p
            
        print("best_blockdim:",best_blockdim,"best_griddim:",best_griddim,"kp:",best_kp,"time:",best_time)
        self.k_p = best_kp
        self.blockdim = best_blockdim
        self.griddim = best_griddim


from inspectnn.Conv.ConvLayer import ConvLayer
from inspectnn.Dense.DenseLayer import DenseLayer
from inspectnn import Dataset
from inspectnn.Pooling.PoolingLayer import PoolingLayer
from inspectnn.Pooling.PoolingLayer_TFLITE import PoolingLayer_TFLITE
from inspectnn.Flatten.FlattenLayer import FlattenLayer
from inspectnn.Model.NetworkModel import NetworkModel
from inspectnn.Model.GenericModel_tflite import GenericModelTflite
from inspectnn.Quantizate.QuantizateLayer import QuantizateLayer
from inspectnn.Add.AddLayer import AddLayer
name = "inspectnn"
__version__ = "0.6.1"
__author__ = "Salvatore Barone, Filippo Ferrandino"
__credits__ = "Department of Electrical Engineering and Information Technologies, University of Naples Federico II, Via Claudio 21, Naples, Italy"
